"""
Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import numpy as np
from numpy.testing import assert_array_equal
from pytest import approx

from eolearn.core import FeatureType
from eolearn.mask import CloudMaskTask


def test_mono_temporal_cloud_detection(test_eopatch):
    add_tcm = CloudMaskTask(
        data_feature=(FeatureType.DATA, "BANDS-S2-L1C"),
        all_bands=True,
        is_data_feature=(FeatureType.MASK, "IS_DATA"),
        mono_features=("CLP_TEST", "CLM_TEST"),
        mask_feature=None,
        average_over=4,
        dilation_size=2,
        mono_threshold=0.4,
    )
    eop_clm = add_tcm(test_eopatch)

    assert_array_equal(eop_clm.mask["CLM_TEST"], eop_clm.mask["CLM_S2C"])
    assert_array_equal(eop_clm.data["CLP_TEST"], eop_clm.data["CLP_S2C"])


def test_multi_temporal_cloud_detection_downscaled(test_eopatch):

    add_tcm = CloudMaskTask(
        data_feature=(FeatureType.DATA, "BANDS-S2-L1C"),
        processing_resolution=120,
        mono_features=("CLP_TEST", "CLM_TEST"),
        multi_features=("CLP_MULTI_TEST", "CLM_MULTI_TEST"),
        mask_feature=(FeatureType.MASK, "CLM_INTERSSIM_TEST"),
        average_over=8,
        dilation_size=4,
    )
    eop_clm = add_tcm(test_eopatch)

    # Check shape and type
    for feature in ((FeatureType.MASK, "CLM_TEST"), (FeatureType.DATA, "CLP_TEST")):
        assert eop_clm[feature].ndim == 4
        assert eop_clm[feature].shape[:-1] == eop_clm.data["BANDS-S2-L1C"].shape[:-1]
        assert eop_clm[feature].shape[-1] == 1
    assert eop_clm.mask["CLM_TEST"].dtype == bool
    assert eop_clm.data["CLP_TEST"].dtype == np.float32

    # Compare mean cloud coverage with provided reference
    assert np.mean(eop_clm.mask["CLM_TEST"]) == approx(np.mean(eop_clm.mask["CLM_S2C"]), abs=0.01)
    assert np.mean(eop_clm.data["CLP_TEST"]) == approx(np.mean(eop_clm.data["CLP_S2C"]), abs=0.01)

    # Check if most of the same times are flagged as cloudless
    cloudless = np.mean(eop_clm.mask["CLM_TEST"], axis=(1, 2, 3)) == 0
    assert np.mean(cloudless == eop_clm.label["IS_CLOUDLESS"][:, 0]) > 0.94

    # Check multi-temporal results and final mask
    assert_array_equal(eop_clm.data["CLP_MULTI_TEST"], eop_clm.data["CLP_MULTI"])
    assert_array_equal(eop_clm.mask["CLM_MULTI_TEST"], eop_clm.mask["CLM_MULTI"])
    assert_array_equal(eop_clm.mask["CLM_INTERSSIM_TEST"], eop_clm.mask["CLM_INTERSSIM"])



"""
Module for creating mask features

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)
Copyright (c) 2018-2019 Johannes Schmid (GeoVille)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
from typing import Callable, Union

import numpy as np

from eolearn.core import EOTask, ZipFeatureTask


class JoinMasksTask(ZipFeatureTask):
    """Joins together masks with the provided logical operation."""

    def __init__(self, input_features, output_feature, join_operation: Union[str, Callable] = "and"):
        """
        :param input_features: Mask features to be joined together.
        :param output_feature: Feature to which to save the joined mask.
        :param join_operation: How to join masks. Supports `'and'`, `'or'`, `'xor'`, or a `Callable` object.
        """
        input_features = self.parse_features(input_features)
        output_feature = self.parse_feature(output_feature)

        if isinstance(join_operation, str):
            methods = {"and": np.logical_and, "or": np.logical_or, "xor": np.logical_xor}
            if join_operation not in methods:
                raise ValueError(
                    f"Join operation {join_operation} is not a viable choice. For operations other than {list(methods)}"
                    "the user must provide a `Callable` object."
                )
            self.join_method = methods[join_operation]
        else:
            self.join_method = join_operation

        super().__init__(input_features, output_feature)

    def zip_method(self, *masks: np.ndarray) -> np.ndarray:
        """Joins masks using the provided operation"""
        final_mask, *masks = masks
        for mask in masks:
            final_mask = self.join_method(final_mask, mask)
        return final_mask


class MaskFeatureTask(EOTask):
    """Masks out values of a feature using defined values of a given mask feature.

    As an example, it can be used to mask the data feature using values from the Sen2cor Scene Classification
    Layer (SCL).

    Contributor: Johannes Schmid, GeoVille Information Systems GmbH, 2018
    """

    def __init__(self, feature, mask_feature, mask_values, no_data_value=np.nan):
        """
        :param feature: A feature to be masked with optional new feature name
        :type feature: (FeatureType, str) or (FeatureType, str, str)
        :param mask_feature: Masking feature. Values of this mask will be used to mask values of `feature`
        :type mask_feature: (FeatureType, str)
        :param mask_values: List of values of `mask_feature` to be used for masking `feature`
        :type mask_values: list of int
        :param no_data_value: Value that replaces masked values in `feature`. Default is `NaN`
        :type no_data_value: float
        :return: The same `eopatch` instance with a masked array
        """
        self.renamed_feature = self.parse_renamed_feature(feature)
        self.mask_feature = self.parse_feature(mask_feature)
        self.mask_values = mask_values
        self.no_data_value = no_data_value

        if not isinstance(self.mask_values, list):
            raise ValueError("Incorrect format or values of argument 'mask_values'")

    def execute(self, eopatch):
        """Mask values of `feature` according to the `mask_values` in `mask_feature`

        :param eopatch: `eopatch` to be processed
        :return: Same `eopatch` instance with masked `feature`
        """
        feature_type, feature_name, new_feature_name = self.renamed_feature
        mask_feature_type, mask_feature_name = self.mask_feature

        data = np.copy(eopatch[feature_type][feature_name])
        mask = eopatch[mask_feature_type][mask_feature_name]

        for value in self.mask_values:
            data = apply_mask(data, mask, value, self.no_data_value, feature_type, mask_feature_type)

        eopatch[feature_type][new_feature_name] = data
        return eopatch


def apply_mask(data, mask, old_value, new_value, data_type, mask_type):
    """A general masking function

    :param data: A data feature
    :type data: numpy.ndarray
    :param mask: A mask feature
    :type mask: numpy.ndarray
    :param old_value: An old value in data that will be replaced
    :type old_value: float
    :param new_value: A new value that will replace the old value in data
    :type new_value: float
    :param data_type: A data feature type
    :type data_type: FeatureType
    :param mask_type: A mask feature type
    :type mask_type: FeatureType
    """
    if not (data_type.is_spatial() and mask_type.is_spatial()):
        raise ValueError("Masking with non-spatial data types is not yet supported")

    if data_type.is_timeless() and mask_type.is_temporal():
        raise ValueError("Cannot mask timeless data feature with time dependent mask feature")

    if data.shape[-3:-1] != mask.shape[-3:-1]:
        raise ValueError("Data feature and mask feature have different spatial dimensions")
    if mask_type.is_temporal() and data.shape[0] != mask.shape[0]:
        raise ValueError("Data feature and mask feature have different temporal dimensions")

    if mask.shape[-1] == data.shape[-1]:
        data[..., mask == old_value] = new_value
    elif mask.shape[-1] == 1:
        data[..., mask[..., 0] == old_value, :] = new_value
    else:
        raise ValueError(
            f"Mask feature has {mask.shape[-1]} number of bands while data feature has {data.shape[-1]} number of bands"
        )
    return data



"""
Module with global fixtures
"""
import os

import pytest

from eolearn.core import EOPatch

EXAMPLE_DATA_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..", "..", "example_data")
TEST_EOPATCH_PATH = os.path.join(EXAMPLE_DATA_PATH, "TestEOPatch")


@pytest.fixture(name="test_eopatch")
def test_eopatch_fixture():
    return EOPatch.load(TEST_EOPATCH_PATH, lazy_loading=True)



"""
Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import numpy as np
import pytest

from eolearn.core import EOPatch, FeatureType
from eolearn.mask import JoinMasksTask, MaskFeatureTask

BANDS_FEATURE = FeatureType.DATA, "BANDS-S2-L1C"
NDVI_FEATURE = FeatureType.DATA, "NDVI"
CLOUD_MASK_FEATURE = FeatureType.MASK, "CLM"
LULC_FEATURE = FeatureType.MASK_TIMELESS, "LULC"


def test_bands_with_clm(test_eopatch):
    ftype, old_name, new_name = FeatureType.DATA, "BANDS-S2-L1C", "BANDS-S2-L1C_MASKED"

    mask_task = MaskFeatureTask([ftype, old_name, new_name], CLOUD_MASK_FEATURE, mask_values=[True], no_data_value=-1)
    eop = mask_task(test_eopatch)

    masked_count = np.count_nonzero(eop[ftype, new_name] == -1)
    clm_count = np.count_nonzero(eop[CLOUD_MASK_FEATURE])
    bands_num = eop[BANDS_FEATURE].shape[-1]
    assert masked_count == clm_count * bands_num


def test_ndvi_with_clm(test_eopatch):
    ftype, old_name, new_name = FeatureType.DATA, "NDVI", "NDVI_MASKED"

    mask_task = MaskFeatureTask([ftype, old_name, new_name], CLOUD_MASK_FEATURE, mask_values=[True])
    eop = mask_task(test_eopatch)

    masked_count = np.count_nonzero(np.isnan(eop[ftype, new_name]))
    clm_count = np.count_nonzero(eop[CLOUD_MASK_FEATURE])
    assert masked_count == clm_count


def test_clm_with_lulc(test_eopatch):
    ftype, old_name, new_name = FeatureType.MASK, "CLM", "CLM_MASKED"

    mask_task = MaskFeatureTask([ftype, old_name, new_name], LULC_FEATURE, mask_values=[2], no_data_value=255)
    eop = mask_task(test_eopatch)

    masked_count = np.count_nonzero(eop[ftype, new_name] == 255)
    lulc_count = np.count_nonzero(eop[LULC_FEATURE] == 2)
    bands_num = eop[CLOUD_MASK_FEATURE].shape[-1]
    time_num = eop[CLOUD_MASK_FEATURE].shape[0]
    assert masked_count == lulc_count * time_num * bands_num


def test_lulc_with_lulc(test_eopatch):
    ftype, old_name, new_name = FeatureType.MASK_TIMELESS, "LULC", "LULC_MASKED"

    mask_task = MaskFeatureTask([ftype, old_name, new_name], LULC_FEATURE, mask_values=[1], no_data_value=100)
    eop = mask_task(test_eopatch)

    masked_count = np.count_nonzero(eop[ftype, new_name] == 100)
    lulc_count = np.count_nonzero(eop[LULC_FEATURE] == 1)
    assert masked_count == lulc_count


def test_wrong_arguments():
    with pytest.raises(ValueError):
        MaskFeatureTask(BANDS_FEATURE, CLOUD_MASK_FEATURE, mask_values=10)


def test_join_masks():
    eopatch = EOPatch()

    mask1 = (FeatureType.MASK_TIMELESS, "Mask1")
    mask_data1 = np.zeros((10, 10, 1), dtype=np.uint8)
    mask_data1[2:5, 2:5] = 1
    eopatch[mask1] = mask_data1

    mask2 = (FeatureType.MASK_TIMELESS, "Mask2")
    mask_data2 = np.zeros((10, 10, 1), dtype=np.uint8)
    mask_data2[0:3, 7:8] = 1
    eopatch[mask2] = mask_data2

    mask3 = (FeatureType.MASK_TIMELESS, "Mask3")
    mask_data3 = np.zeros((10, 10, 1), dtype=np.uint8)
    mask_data3[1:1] = 1
    eopatch[mask3] = mask_data3

    input_features = [mask1, mask2, mask3]
    output_feature = (FeatureType.MASK_TIMELESS, "Output")

    task1 = JoinMasksTask(input_features, output_feature)
    expected_result = mask_data1 & mask_data2 & mask_data3
    task1(eopatch)
    assert np.array_equal(eopatch[output_feature], expected_result)

    task2 = JoinMasksTask(input_features, output_feature, "or")
    expected_result = mask_data1 | mask_data2 | mask_data3
    task2(eopatch)
    assert np.array_equal(eopatch[output_feature], expected_result)

    task3 = JoinMasksTask(input_features, output_feature, lambda x, y: x + y)
    expected_result = mask_data1 + mask_data2 + mask_data3
    task3(eopatch)
    assert np.array_equal(eopatch[output_feature], expected_result)



"""
Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja, Eva Erzin (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import numpy as np
import pytest

from eolearn.core import FeatureType
from eolearn.mask import SnowMaskTask, TheiaSnowMaskTask


@pytest.mark.parametrize("params", [{"dem_params": (100, 100, 100)}, {"red_params": 45}, {"ndsi_params": (0.2, 3)}])
def test_raises_errors(params, test_eopatch):
    with pytest.raises(ValueError):
        theia_mask = TheiaSnowMaskTask(
            (FeatureType.DATA, "BANDS-S2-L1C"),
            [2, 3, 11],
            (FeatureType.MASK, "CLM"),
            (FeatureType.DATA_TIMELESS, "DEM"),
            **params
        )
        theia_mask(test_eopatch)


@pytest.mark.parametrize(
    "task, result",
    [
        (SnowMaskTask((FeatureType.DATA, "BANDS-S2-L1C"), [2, 3, 7, 11], mask_name="TEST_SNOW_MASK"), (50468, 1405)),
        (
            TheiaSnowMaskTask(
                (FeatureType.DATA, "BANDS-S2-L1C"),
                [2, 3, 11],
                (FeatureType.MASK, "CLM"),
                (FeatureType.DATA_TIMELESS, "DEM"),
                b10_index=10,
                mask_name="TEST_THEIA_SNOW_MASK",
            ),
            (60682, 10088),
        ),
    ],
)
def test_snow_coverage(task, result, test_eopatch):
    resulting_eopatch = task(test_eopatch)
    output = resulting_eopatch[task.mask_feature]

    assert output.ndim == 4
    assert output.shape[:-1] == test_eopatch.data["BANDS-S2-L1C"].shape[:-1]
    assert output.shape[-1] == 1

    assert output.dtype == bool

    snow_pixels = np.sum(output, axis=(1, 2, 3))
    assert np.sum(snow_pixels) == result[0], "Sum of snowy pixels does not match"
    assert snow_pixels[-4] == result[1], "Snowy pixels on specified frame do not match"



"""
Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import numpy as np
import pytest

from eolearn.core import EOPatch, FeatureType
from eolearn.mask import ClassFrequencyTask

IN_FEATURE = (FeatureType.MASK, "TEST")
OUT_FEATURE = (FeatureType.DATA_TIMELESS, "FREQ")


@pytest.mark.parametrize("classes, no_data_value", ((["a", "b"], 0), (4, 0), (None, 0), ([1, 2, 3], 2)))
def test_value_error(classes, no_data_value):
    with pytest.raises(ValueError):
        ClassFrequencyTask(IN_FEATURE, OUT_FEATURE, classes=classes, no_data_value=no_data_value)


def test_class_frequency():
    shape = (20, 5, 5, 2)

    data = np.random.randint(0, 5, size=shape)
    data[:, 0, 0, 0] = 0
    data[:, 0, 1, 0] = 2

    eopatch = EOPatch()
    eopatch[IN_FEATURE] = data

    eopatch = ClassFrequencyTask(IN_FEATURE, OUT_FEATURE, [2, 1, 55])(eopatch)
    result = eopatch[OUT_FEATURE]

    # all zeros through the temporal dimension should result in nan
    assert np.isnan(result[0, 0, 0])

    # frequency of 2 should be 1
    assert result[0, 1, 0] == 1

    # frequency of 55 should be 0
    assert np.all(result[1:, :, -2] == 0)

    assert result.shape == (5, 5, 6)



"""
Public classes and functions of mask subpackage
"""

from .cloud_mask import CloudMaskTask
from .mask_counting import ClassFrequencyTask
from .masking import JoinMasksTask, MaskFeatureTask
from .snow_mask import SnowMaskTask, TheiaSnowMaskTask
from .utils import resize_images

__version__ = "1.0.2"
